import os
from datetime import datetime
from dateutil.tz import tzutc
from unittest.mock import patch

import pyinaturalist
from pyinaturalist.constants import API_V1_BASE_URL
from pyinaturalist.node_api import get_geojson_observations
from pyinaturalist.v1 import get_observation, get_user_by_id, get_users_autocomplete
from test.conftest import load_sample_data


def test_get_user_by_id(requests_mock):
    user_id = 1
    requests_mock.get(
        f'{API_V1_BASE_URL}/users/{user_id}',
        json=load_sample_data('get_user_by_id.json'),
        status_code=200,
    )

    user = get_user_by_id(user_id)
    assert user['id'] == user_id
    assert user['created_at'] == datetime(2008, 3, 20, 21, 15, 42, tzinfo=tzutc())
    assert user['spam'] is False


def test_get_users_autocomplete(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/users/autocomplete',
        json=load_sample_data('get_users_autocomplete.json'),
        status_code=200,
    )

    response = get_users_autocomplete(q='niconoe')
    first_result = response['results'][0]

    assert len(response['results']) == response['total_results'] == 3
    assert first_result['id'] == 886482
    assert first_result['login'] == 'niconoe'
    assert first_result['created_at'] == datetime(2018, 4, 23, 17, 11, 14, tzinfo=tzutc())


# TODO: Deprecate get_geojson_observations and move to pyinaturalist-convert
def test_get_geojson_observations(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_observation.json'),
        status_code=200,
    )

    geojson = get_geojson_observations(id=16227955)
    feature = geojson['features'][0]
    assert feature['geometry']['coordinates'] == [4.360086, 50.646894]
    assert feature['properties']['id'] == 16227955
    assert feature['properties']['taxon_id'] == 493595


def test_get_geojson_observations__custom_properties(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_observation.json'),
        status_code=200,
    )

    properties = ['taxon_name', 'taxon_rank']
    geojson = get_geojson_observations(id=16227955, properties=properties)
    feature = geojson['features'][0]
    assert feature['properties']['taxon_name'] == 'Lixus bardanae'
    assert feature['properties']['taxon_rank'] == 'species'
    assert 'id' not in feature['properties'] and 'taxon_id' not in feature['properties']


# TODO: Deprecate globals in favort of API client settings
@patch.dict(os.environ, {'GITHUB_REF': 'refs/heads/main'}, clear=True)
def test_user_agent(requests_mock):
    requests_mock.get(
        f'{API_V1_BASE_URL}/observations',
        json=load_sample_data('get_observation.json'),
        status_code=200,
    )
    default_ua = pyinaturalist.DEFAULT_USER_AGENT

    # By default, we have a 'Pyinaturalist' user agent:
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == default_ua

    # But if the user sets a custom one, it is indeed used:
    get_observation(16227955, user_agent='CustomUA')
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'CustomUA'

    # We can also set it globally:
    pyinaturalist.user_agent = 'GlobalUA'
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'GlobalUA'

    # And it persists across requests:
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'GlobalUA'

    # But if we have a global and local one, the local has priority
    get_observation(16227955, user_agent='CustomUA 2')
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == 'CustomUA 2'

    # We can reset the global settings to the default:
    pyinaturalist.user_agent = pyinaturalist.DEFAULT_USER_AGENT
    get_observation(16227955)
    assert requests_mock._adapter.last_request._request.headers['User-Agent'] == default_ua



from datetime import date, datetime
from typing import List, Union

from attr import field

from pyinaturalist.constants import TableRow
from pyinaturalist.converters import safe_split, try_int_or_float
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    Taxon,
    User,
    datetime_now_attr,
    define_model,
    kwarg,
)

# Mappings from observation field value datatypes to python datatypes
OFV_DATATYPES = {
    'dna': str,
    'date': date,
    'datetime': datetime,
    'numeric': try_int_or_float,
    'taxon': int,
    'text': str,
    'time': str,
}
OFVValue = Union[date, datetime, float, int, str]


@define_model
class ObservationField(BaseModel):
    """An observation field **definition**, based on the schema of
    `GET /observation_fields <https://www.inaturalist.org/pages/api+reference#get-observation_fields>`_.
    """

    allowed_values: List[str] = field(converter=safe_split, factory=list)
    created_at: datetime = datetime_now_attr
    datatype: str = kwarg  # Enum
    description: str = kwarg
    id: int = kwarg
    name: str = kwarg
    updated_at: datetime = datetime_now_attr
    user_id: int = kwarg
    users_count: int = kwarg
    uuid: str = kwarg
    values_count: int = kwarg

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Type': self.datatype,
            'Name': self.name,
            'Description': self.description,
        }

    def __str__(self) -> str:
        description = ': {self.description}' if self.description else ''
        return f'[{self.id}] {self.name} ({self.datatype}){description}'


@define_model
class ObservationFieldValue(BaseModel):
    """An observation field **value**, matching the schema of ``ofvs``
    from `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    datatype: str = kwarg  # Enum
    field_id: int = kwarg
    id: int = kwarg
    name: str = kwarg
    taxon_id: int = kwarg
    user_id: int = kwarg
    uuid: str = kwarg
    value: OFVValue = kwarg

    # Lazy-loaded nested model objects
    taxon: property = LazyProperty(Taxon.from_json)
    user: property = LazyProperty(User.from_json)

    # Unused attrbiutes
    # name_ci: str = kwarg
    # value_ci: int = kwarg

    # Convert value by datatype
    def __attrs_post_init__(self):
        if self.datatype in OFV_DATATYPES and self.value is not None:
            converter = OFV_DATATYPES[self.datatype]
            self.value = converter(self.value)

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Type': self.datatype,
            'Name': self.name,
            'Value': self.value,
        }

    def __str__(self) -> str:
        return f'{self.name}: {self.value}'



from datetime import datetime
from dateutil.tz import tzutc

{
    'id': 1,
    'login': 'kueda',
    'spam': False,
    'suspended': False,
    'created_at': datetime(2008, 3, 20, 21, 15, 42, tzinfo=tzutc()),
    'login_autocomplete': 'kueda',
    'login_exact': 'kueda',
    'name': 'Ken-ichi Ueda',
    'name_autocomplete': 'Ken-ichi Ueda',
    'orcid': 'https://orcid.org/0000-0003-0145-6846',
    'icon': 'https://static.inaturalist.org/attachments/users/icons/1/thumb.jpg?1475527316',
    'observations_count': 40947,
    'identifications_count': 95624,
    'journal_posts_count': 86,
    'activity_count': 136657,
    'species_count': 8831,
    'universal_search_rank': 40947,
    'roles': ['admin', 'curator'],
    'site_id': 1,
    'icon_url': 'https://static.inaturalist.org/attachments/users/icons/1/medium.jpg?1475527316',
}



"""
Functions to access the original (v0) iNaturalist API
See: https://www.inaturalist.org/pages/api+reference
"""
# flake8: noqa: F401, F403

from pyinaturalist.v0.observation_fields import get_observation_fields, put_observation_field_values
from pyinaturalist.v0.observations import (
    create_observation,
    delete_observation,
    get_observations,
    update_observation,
    upload_photos,
    upload_sounds,
)



import os
import pytest
from unittest.mock import patch

from keyring.errors import KeyringError
from requests import HTTPError

from pyinaturalist.auth import get_access_token, get_keyring_credentials, set_keyring_credentials
from pyinaturalist.constants import API_V0_BASE_URL
from pyinaturalist.exceptions import AuthenticationError
from test.conftest import MOCK_CREDS_ENV, MOCK_CREDS_OAUTH, load_sample_data

token_accepted_json = load_sample_data('get_access_token.json')
token_rejected_json = load_sample_data('get_access_token_401.json')


# get_access_token
# --------------------


@patch.dict(os.environ, {}, clear=True)
def test_get_access_token(requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/oauth/token',
        json=token_accepted_json,
        status_code=200,
    )

    token = get_access_token('valid_username', 'valid_password', 'valid_app_id', 'valid_app_secret')
    assert token == '604e5df329b98eecd22bb0a84f88b68'


@patch.dict(os.environ, MOCK_CREDS_ENV)
@patch('pyinaturalist.auth.get_keyring_credentials')
def test_get_access_token__envars(mock_keyring_credentials, requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/oauth/token',
        json=token_accepted_json,
        status_code=200,
    )

    token = get_access_token()
    assert token == '604e5df329b98eecd22bb0a84f88b68'
    mock_keyring_credentials.assert_not_called()


@patch.dict(
    os.environ,
    {
        'INAT_USERNAME': 'valid_username',
        'INAT_PASSWORD': 'valid_password',
    },
)
def test_get_access_token__mixed_args_and_envars(requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/oauth/token',
        json=token_accepted_json,
        status_code=200,
    )

    token = get_access_token(app_id='valid_app_id', app_secret='valid_app_secret')
    assert token == '604e5df329b98eecd22bb0a84f88b68'


@patch.dict(os.environ, {}, clear=True)
@patch('pyinaturalist.auth.get_keyring_credentials', return_value=MOCK_CREDS_OAUTH)
@patch('pyinaturalist.auth.post')
def test_get_access_token__keyring(mock_post, mock_keyring_credentials, requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/oauth/token',
        json=token_accepted_json,
        status_code=200,
    )

    get_access_token()
    submitted_json = mock_post.call_args[1]['json']
    assert submitted_json == {'grant_type': 'password', **MOCK_CREDS_OAUTH}
    mock_keyring_credentials.assert_called()


@patch.dict(os.environ, {}, clear=True)
@patch('pyinaturalist.auth.get_keyring_credentials')
def test_get_access_token__missing_creds(mock_keyring_credentials):
    with pytest.raises(AuthenticationError):
        get_access_token('username')


def test_get_access_token__invalid_creds(requests_mock):
    requests_mock.post(
        f'{API_V0_BASE_URL}/oauth/token',
        json=token_rejected_json,
        status_code=401,
    )

    with pytest.raises(HTTPError):
        get_access_token('username', 'password', 'app_id', 'app_secret')


# get_keyring_credentials
# -------------------------


@patch('keyring.get_password', side_effect=list(MOCK_CREDS_OAUTH.values()))
def test_get_keyring_credentials(get_password):
    assert get_keyring_credentials() == MOCK_CREDS_OAUTH


def test_get_keyring_credentials__not_installed():
    pass


@patch('keyring.get_password', side_effect=KeyringError)
def test_get_keyring_credentials__no_backend(get_password):
    assert get_keyring_credentials() == {}


# get_keyring_credentials
# -------------------------


@patch('keyring.set_password')
def test_set_keyring_credentials(set_password):
    set_keyring_credentials('username', 'password', 'app_id', 'app_secret')
    assert set_password.call_count == 4



"""Some common functions for HTTP requests used by all API modules"""
import threading
from contextlib import contextmanager
from logging import getLogger
from os import getenv
from typing import Dict
from unittest.mock import Mock

from requests import Response, Session

import pyinaturalist
from pyinaturalist.api_docs import copy_signature
from pyinaturalist.constants import (
    MAX_DELAY,
    REQUESTS_PER_DAY,
    REQUESTS_PER_MINUTE,
    REQUESTS_PER_SECOND,
    WRITE_HTTP_METHODS,
    MultiInt,
    RequestParams,
)
from pyinaturalist.request_params import prepare_request

# Mock response content to return in dry-run mode
MOCK_RESPONSE = Mock(spec=Response)
MOCK_RESPONSE.json.return_value = {'results': [], 'total_results': 0, 'access_token': ''}

logger = getLogger(__name__)
thread_local = threading.local()


def request(
    method: str,
    url: str,
    access_token: str = None,
    user_agent: str = None,
    ids: MultiInt = None,
    headers: Dict = None,
    json: Dict = None,
    session: Session = None,
    raise_for_status: bool = True,
    timeout: float = 5,
    **params: RequestParams,
) -> Response:
    """Wrapper around :py:func:`requests.request` that supports dry-run mode and rate-limiting,
    and adds appropriate headers.

    Args:
        method: HTTP method
        url: Request URL
        access_token: access_token: the access token, as returned by :func:`get_access_token()`
        user_agent: A custom user-agent string to provide to the iNaturalist API
        ids: One or more integer IDs used as REST resource(s) to request
        headers: Request headers
        json: JSON request body
        session: Existing Session object to use instead of creating a new one
        timeout: Time (in seconds) to wait for a response from the server; if exceeded, a
            :py:exc:`requests.exceptions.Timeout` will be raised.
        params: All other keyword arguments are interpreted as request parameters

    Returns:
        API response
    """
    url, params, headers, json = prepare_request(
        url,
        access_token,
        user_agent,
        ids,
        params,
        headers,
        json,
    )
    session = session or get_session()

    # Run either real request or mock request depending on settings
    if is_dry_run_enabled(method):
        logger.debug('Dry-run mode enabled; mocking request')
        log_request(method, url, params=params, headers=headers)
        return MOCK_RESPONSE
    else:
        with ratelimit():
            response = session.request(
                method, url, params=params, headers=headers, json=json, timeout=timeout
            )
        if raise_for_status:
            response.raise_for_status()
        return response


@copy_signature(request, exclude='method')
def delete(url: str, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.delete` that supports dry-run mode and rate-limiting"""
    return request('DELETE', url, **kwargs)


@copy_signature(request, exclude='method')
def get(url: str, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.get` that supports dry-run mode and rate-limiting"""
    return request('GET', url, **kwargs)


@copy_signature(request, exclude='method')
def post(url: str, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.post` that supports dry-run mode and rate-limiting"""
    return request('POST', url, **kwargs)


@copy_signature(request, exclude='method')
def put(url: str, **kwargs) -> Response:
    """Wrapper around :py:func:`requests.put` that supports dry-run mode and rate-limiting"""
    return request('PUT', url, **kwargs)


# TODO: Handle error 429 if we still somehow exceed the rate limit?
@contextmanager
def ratelimit(bucket=pyinaturalist.user_agent):
    """Add delays in between requests to stay within the rate limits. If pyrate-limiter is
    not installed, this will quietly do nothing.
    """
    if RATE_LIMITER:
        with RATE_LIMITER.ratelimit(bucket, delay=True, max_delay=MAX_DELAY):
            yield
    else:
        yield


def get_limiter():
    """Get a rate limiter object, if pyrate-limiter is installed"""
    try:
        from pyrate_limiter import Duration, Limiter, RequestRate

        requst_rates = [
            RequestRate(REQUESTS_PER_SECOND, Duration.SECOND),
            RequestRate(REQUESTS_PER_MINUTE, Duration.MINUTE),
            RequestRate(REQUESTS_PER_DAY, Duration.DAY),
        ]
        return Limiter(*requst_rates)
    except ImportError:
        return None


def get_session() -> Session:
    """Get a Session object that will be reused across requests to take advantage of connection
    pooling. This is especially relevant for large paginated requests. If used in a multi-threaded
    context (for example, a :py:class:`~concurrent.futures.ThreadPoolExecutor`), a separate session
    is used for each thread.
    """
    if not hasattr(thread_local, "session"):
        thread_local.session = Session()
    return thread_local.session


def is_dry_run_enabled(method: str) -> bool:
    """A wrapper to determine if dry-run (aka test mode) has been enabled via either
    a constant or an environment variable. Dry-run mode may be enabled for either write
    requests, or all requests.
    """
    dry_run_enabled = pyinaturalist.DRY_RUN_ENABLED or env_to_bool('DRY_RUN_ENABLED')
    if method in WRITE_HTTP_METHODS:
        return dry_run_enabled or pyinaturalist.DRY_RUN_WRITE_ONLY or env_to_bool('DRY_RUN_WRITE_ONLY')
    return dry_run_enabled


def env_to_bool(environment_variable: str) -> bool:
    """Translate an environment variable to a boolean value, accounting for minor
    variations (case, None vs. False, etc.)
    """
    env_value = getenv(environment_variable)
    return bool(env_value) and str(env_value).lower() not in ['false', 'none']


def log_request(*args, **kwargs):
    """Log all relevant information about an HTTP request"""
    kwargs_strs = [f'{k}={v}' for k, v in kwargs.items()]
    logger.info('Request: {}'.format(', '.join(list(args) + kwargs_strs)))


RATE_LIMITER = get_limiter()



[
    "   52747:       Family Vespidae (Hornets, Paper Wasps, Potter Wasps, and Allies)",
    "   84738:    Subfamily Vespinae (Hornets and Yellowjackets)",
    "  131878:      Species Nicrophorus vespillo (Vespillo Burying Beetle)",
    "  621585:      Species Vespicula trachinoides (Vespicula Waspfish)",
    "  495392:      Species Vespidae st1",
    "   70118:      Species Nicrophorus vespilloides (Lesser Vespillo Burying Beetle)",
    "   92786:        Genus Vespicula",
    "  646195:        Genus Vespiodes",
    "  621584:      Species Vespicula cypho",
    "  621586:      Species Vespicula zollingeri",
]



"""Dataclasses for modeling iNaturalist API response objects"""
# flake8: noqa: F401
# TODO: Validators for multiple choice fields?
from typing import Callable
from attr import define

from pyinaturalist.models.base import (
    ModelObjects,
    ResponseOrObject,
    ResponseOrObjects,
    BaseModel,
    BaseModelCollection,
    datetime_attr,
    datetime_now_attr,
    coordinate_pair,
    kwarg,
    load_json,
)
from pyinaturalist.models.lazy_property import LazyProperty, add_lazy_attrs, get_model_fields

# attrs class decorator with the most commonly used options
define_model: Callable = define(auto_attribs=False, field_transformer=add_lazy_attrs)

# Imported in order of model dependencies
from pyinaturalist.models.photo import Photo
from pyinaturalist.models.place import Place
from pyinaturalist.models.user import User
from pyinaturalist.models.taxon import (
    ConservationStatus,
    EstablishmentMeans,
    Taxon,
    TaxonCount,
    TaxonCounts,
)
from pyinaturalist.models.controlled_term import Annotation, ControlledTerm, ControlledTermValue
from pyinaturalist.models.comment import Comment
from pyinaturalist.models.identification import Identification
from pyinaturalist.models.life_list import LifeList, LifeListTaxon
from pyinaturalist.models.project import (
    Project,
    ProjectObservation,
    ProjectObservationField,
    ProjectUser,
)
from pyinaturalist.models.observation_field import ObservationField, ObservationFieldValue
from pyinaturalist.models.observation import Observation
from pyinaturalist.models.search import SearchResult


# Add aliases for some of the longer class names
ID = Identification
OFD = ObservationField
OFV = ObservationFieldValue



from datetime import datetime
from typing import List

from attr import field

from pyinaturalist.constants import INAT_BASE_URL, TableRow
from pyinaturalist.models import BaseModel, datetime_now_attr, define_model, kwarg


@define_model
class User(BaseModel):
    """A dataclass containing information about an user, matching the schema of
    `GET /users/{id} <https://api.inaturalist.org/v1/docs/#!/Users/get_users_id>`_.
    """

    activity_count: int = kwarg
    created_at: datetime = datetime_now_attr
    icon: str = kwarg
    icon_url: str = kwarg
    id: int = kwarg
    identifications_count: int = kwarg
    journal_posts_count: int = kwarg
    login: str = kwarg
    name: str = kwarg
    observations_count: int = kwarg
    orcid: str = kwarg
    roles: List = field(factory=list)
    site_id: int = kwarg

    # Unused attributes
    # login_autocomplete: str = kwarg
    # login_exact: str = kwarg
    # name_autocomplete: str = kwarg
    # spam: bool = kwarg
    # suspended: bool = kwarg
    # universal_search_rank: int = kwarg

    # Aliases
    @property
    def username(self) -> str:
        return self.login

    @property
    def display_name(self) -> str:
        return self.name

    @property
    def url(self) -> str:
        """Info URL on iNaturalist.org"""
        return f'{INAT_BASE_URL}/users/{self.id}'

    @property
    def row(self) -> TableRow:
        return {
            'ID': self.id,
            'Username': self.username,
            'Display name': self.display_name,
            'Obs. count': self.observations_count,
            'ID count': self.identifications_count,
        }

    def __str__(self) -> str:
        real_name = f' ({self.name})' if self.name else ''
        return f"[{self.id}] {self.login}{real_name}"



from datetime import datetime
from dateutil.tz import tzoffset

{
    'total_results': 873,
    'page': 1,
    'per_page': 2,
    'results': [
        {
            'id': 155554373,
            'uuid': '88031477-a183-46c6-8de3-06bac7b5f220',
            'user': {
                'id': 2115051,
                'login': 'jkcook',
                'spam': False,
                'suspended': False,
                'created_at': '2019-08-10T17:52:01+00:00',
                'login_autocomplete': 'jkcook',
                'login_exact': 'jkcook',
                'name': 'Jordan Cook',
                'name_autocomplete': 'Jordan Cook',
                'orcid': None,
                'icon': 'https://static.inaturalist.org/attachments/users/icons/2115051/thumb.jpeg?1591994381',
                'observations_count': 706,
                'identifications_count': 542,
                'journal_posts_count': 0,
                'activity_count': 1248,
                'species_count': 357,
                'universal_search_rank': 706,
                'roles': [],
                'site_id': 1,
                'icon_url': 'https://static.inaturalist.org/attachments/users/icons/2115051/medium.jpeg?1591994381',
            },
            'created_at': datetime(2021, 2, 18, 20, 31, 32, tzinfo=tzoffset(None, -21600)),
            'created_at_details': {
                'date': '2021-02-18',
                'day': 18,
                'month': 2,
                'year': 2021,
                'hour': 20,
                'week': 7,
            },
            'body': None,
            'category': 'supporting',
            'current': True,
            'flags': [],
            'own_observation': True,
            'taxon_change': None,
            'vision': False,
            'disagreement': False,
            'previous_observation_taxon_id': 60132,
            'spam': False,
            'taxon_id': 60132,
            'hidden': False,
            'current_taxon': True,
            'taxon': {
                'id': 60132,
                'rank': 'species',
                'rank_level': 10,
                'iconic_taxon_id': 47126,
                'ancestor_ids': [
                    47126,
                    211194,
                    47125,
                    47124,
                    47605,
                    47604,
                    201654,
                    632883,
                    632885,
                    48561,
                    60132,
                ],
                'is_active': True,
                'min_species_taxon_id': 60132,
                'min_species_ancestry': '47126,211194,47125,47124,47605,47604,201654,632883,632885,48561,60132',
            },
            'observation': 'TRUNCATED',
            'moderator_actions': [],
            'observed_on': None,
        },
        {
            'id': 155554077,
            'uuid': '12922da1-a203-46f2-b0f2-fd21917b4b07',
            'user': {
                'id': 2115051,
                'login': 'jkcook',
                'spam': False,
                'suspended': False,
                'created_at': '2019-08-10T17:52:01+00:00',
                'login_autocomplete': 'jkcook',
                'login_exact': 'jkcook',
                'name': 'Jordan Cook',
                'name_autocomplete': 'Jordan Cook',
                'orcid': None,
                'icon': 'https://static.inaturalist.org/attachments/users/icons/2115051/thumb.jpeg?1591994381',
                'observations_count': 706,
                'identifications_count': 542,
                'journal_posts_count': 0,
                'activity_count': 1248,
                'species_count': 357,
                'universal_search_rank': 706,
                'roles': [],
                'site_id': 1,
                'icon_url': 'https://static.inaturalist.org/attachments/users/icons/2115051/medium.jpeg?1591994381',
            },
            'created_at': datetime(2021, 2, 18, 20, 29, 6, tzinfo=tzoffset(None, -21600)),
            'created_at_details': {
                'date': '2021-02-18',
                'day': 18,
                'month': 2,
                'year': 2021,
                'hour': 20,
                'week': 7,
            },
            'body': None,
            'category': 'supporting',
            'current': True,
            'flags': [],
            'own_observation': True,
            'taxon_change': None,
            'vision': False,
            'disagreement': False,
            'previous_observation_taxon_id': 61340,
            'spam': False,
            'taxon_id': 61340,
            'hidden': False,
            'current_taxon': True,
            'taxon': {
                'id': 61340,
                'rank': 'species',
                'rank_level': 10,
                'iconic_taxon_id': 47126,
                'ancestor_ids': [
                    47126,
                    211194,
                    47125,
                    47124,
                    47605,
                    47604,
                    201654,
                    632883,
                    632885,
                    48561,
                    61340,
                ],
                'is_active': True,
                'min_species_taxon_id': 61340,
                'min_species_ancestry': '47126,211194,47125,47124,47605,47604,201654,632883,632885,48561,61340',
            },
            'observation': 'TRUNCATED',
            'moderator_actions': [],
            'observed_on': None,
        },
    ],
}



import pytest
from datetime import date, datetime
from dateutil.tz import gettz
from unittest.mock import patch

from pyinaturalist.request_params import (
    convert_bool_params,
    convert_datetime_params,
    convert_list_params,
    convert_observation_fields,
    convert_pagination_params,
    get_interval_ranges,
    preprocess_request_body,
    preprocess_request_params,
    strip_empty_values,
    validate_ids,
    validate_multiple_choice_param,
    validate_multiple_choice_params,
)

TEST_PARAMS = {
    'is_active': False,
    'only_id': 'true',
    'preferred_place_id': [1, 2],
    'rank': ['phylum', 'class'],
    'q': '',
    'locale': None,
    'parent_id': 1,
    'observation_fields': {1: 'value'},
}


def test_convert_bool_params():
    params = convert_bool_params(TEST_PARAMS)
    assert params['is_active'] == 'false'
    assert params['only_id'] == 'true'


# Test some recognized date(time) formats, with and without TZ info, in date and non-date params
@pytest.mark.parametrize(
    'param, value, expected',
    [
        ('created_d1', '19951231T235959', '1995-12-31T23:59:59-08:00'),
        ('created_d2', '2008-08-08 08:08:08Z', '2008-08-08T08:08:08+00:00'),
        ('created_on', '2010-10-10 10:10:10-05:00', '2010-10-10T10:10:10-05:00'),
        ('created_on', 'Jan 1 2000', '2000-01-01T00:00:00-08:00'),
        ('d1', '19970716', '1997-07-16T00:00:00-07:00'),
        ('q', date(1954, 2, 5), '1954-02-05'),
        ('q', datetime(1954, 2, 5), '1954-02-05T00:00:00-08:00'),
        ('q', 'not a datetime', 'not a datetime'),
    ],
)
@patch('pyinaturalist.converters.tzlocal', return_value=gettz('US/Pacific'))
def test_convert_datetime_params(tzlocal, param, value, expected):
    converted = convert_datetime_params({param: value})
    assert converted[param] == expected


# Test both int and string lists
def test_convert_list_params():
    params = convert_list_params(TEST_PARAMS)
    assert params['preferred_place_id'] == '1,2'
    assert params['rank'] == 'phylum,class'


def test_convert_observation_fields():
    params = convert_observation_fields(TEST_PARAMS)
    assert params['observation_field_values_attributes'] == [
        {'observation_field_id': 1, 'value': 'value'}
    ]


def test_convert_pagination_params():
    params = convert_pagination_params({'per_page': 100})
    assert params['per_page'] == 100

    params = convert_pagination_params({'per_page': 100, 'count_only': True})
    assert params['per_page'] == 0
    assert 'count_only' not in params

    params = convert_pagination_params({'per_page': 100, 'count_only': False})
    assert params['per_page'] == 100
    assert 'count_only' not in params


def test_get_interval_ranges__monthly():
    expected_ranges = [
        (datetime(2020, 1, 1, 0, 0), datetime(2020, 1, 31, 0, 0)),
        (datetime(2020, 2, 1, 0, 0), datetime(2020, 2, 29, 0, 0)),
        (datetime(2020, 3, 1, 0, 0), datetime(2020, 3, 31, 0, 0)),
        (datetime(2020, 4, 1, 0, 0), datetime(2020, 4, 30, 0, 0)),
        (datetime(2020, 5, 1, 0, 0), datetime(2020, 5, 31, 0, 0)),
        (datetime(2020, 6, 1, 0, 0), datetime(2020, 6, 30, 0, 0)),
        (datetime(2020, 7, 1, 0, 0), datetime(2020, 7, 31, 0, 0)),
        (datetime(2020, 8, 1, 0, 0), datetime(2020, 8, 31, 0, 0)),
        (datetime(2020, 9, 1, 0, 0), datetime(2020, 9, 30, 0, 0)),
        (datetime(2020, 10, 1, 0, 0), datetime(2020, 10, 31, 0, 0)),
        (datetime(2020, 11, 1, 0, 0), datetime(2020, 11, 30, 0, 0)),
        (datetime(2020, 12, 1, 0, 0), datetime(2020, 12, 31, 0, 0)),
    ]
    ranges = get_interval_ranges(datetime(2020, 1, 1), datetime(2020, 12, 31), 'monthly')
    assert ranges == expected_ranges


def test_get_interval_ranges__yearly():
    expected_ranges = [
        (datetime(2010, 1, 1, 0, 0), datetime(2010, 12, 31, 0, 0)),
        (datetime(2011, 1, 1, 0, 0), datetime(2011, 12, 31, 0, 0)),
        (datetime(2012, 1, 1, 0, 0), datetime(2012, 12, 31, 0, 0)),
        (datetime(2013, 1, 1, 0, 0), datetime(2013, 12, 31, 0, 0)),
        (datetime(2014, 1, 1, 0, 0), datetime(2014, 12, 31, 0, 0)),
        (datetime(2015, 1, 1, 0, 0), datetime(2015, 12, 31, 0, 0)),
        (datetime(2016, 1, 1, 0, 0), datetime(2016, 12, 31, 0, 0)),
        (datetime(2017, 1, 1, 0, 0), datetime(2017, 12, 31, 0, 0)),
        (datetime(2018, 1, 1, 0, 0), datetime(2018, 12, 31, 0, 0)),
        (datetime(2019, 1, 1, 0, 0), datetime(2019, 12, 31, 0, 0)),
        (datetime(2020, 1, 1, 0, 0), datetime(2020, 12, 31, 0, 0)),
    ]
    ranges = get_interval_ranges(datetime(2010, 1, 1), datetime(2020, 1, 1), 'yearly')
    assert ranges == expected_ranges


def test_get_interval_ranges__invalid():
    with pytest.raises(ValueError):
        get_interval_ranges(datetime(2020, 1, 1), datetime(2020, 12, 31), 'daily')


def test_strip_empty_params():
    params = strip_empty_values(TEST_PARAMS)
    assert len(params) == 6
    assert 'q' not in params and 'locale' not in params
    assert 'is_active' in params and 'only_id' in params


@pytest.mark.parametrize(
    'value, expected',
    [
        ('1', '1'),
        (1, '1'),
        ('1,2,3', '1,2,3'),
        ([1, 2, 3], '1,2,3'),
        ([1e5, 2e5], '100000,200000'),
    ],
)
def test_validate_ids(value, expected):
    assert validate_ids(value) == expected


def test_validate_ids__invalid():
    with pytest.raises(ValueError):
        validate_ids('not a number')


# This is just here so that tests will fail if one of the conversion steps is removed
@patch('pyinaturalist.request_params.convert_bool_params')
@patch('pyinaturalist.request_params.convert_datetime_params')
@patch('pyinaturalist.request_params.convert_list_params')
@patch('pyinaturalist.request_params.strip_empty_values')
def test_preprocess_request_params(mock_bool, mock_datetime, mock_list, mock_strip):
    preprocess_request_params({'id': 1})
    assert all([mock_bool.called, mock_datetime.called, mock_list.called, mock_strip.called])


@patch('pyinaturalist.request_params.convert_bool_params')
@patch('pyinaturalist.request_params.convert_datetime_params')
@patch('pyinaturalist.request_params.convert_list_params')
@patch('pyinaturalist.request_params.strip_empty_values')
def test_preprocess_request_body(mock_bool, mock_datetime, mock_list, mock_strip):
    preprocess_request_body({'id': 1})
    assert all([mock_bool.called, mock_datetime.called, mock_list.called, mock_strip.called])


def test_validate_multiple_choice_param():
    params = {
        'param1': 'valid_str',
        'param2': 'invalid_str',
    }
    choices = ['str1', 'str2', 'valid_str']

    validated_params = validate_multiple_choice_param(params, 'param1', choices)
    assert params == validated_params
    with pytest.raises(ValueError):
        validate_multiple_choice_param(params, 'param2', choices)


@pytest.mark.parametrize(
    'params',
    [
        {'csi': 'LC'},
        {'csi': ['EW', 'EX']},
        {'geoprivacy': 'open'},
        {'iconic_taxa': 'Animalia'},
        {'identifications': 'most_agree'},
        {'license': 'CC-BY-NC'},
        {'rank': 'order'},
        {'quality_grade': 'research'},
        {'search_on': 'tags'},
        {'geoprivacy': ['open', 'obscured']},
        {'geoprivacy': 'open', 'iconic_taxa': 'Animalia', 'license': 'CC-BY-NC'},
    ],
)
def test_validate_multiple_choice_params(params):
    # If valid, no exception should not be raised
    validate_multiple_choice_params(params)

    # If invalid, an exception should be raised
    with pytest.raises(ValueError):
        validate_multiple_choice_params({k: 'Invalid_value' for k in params})

    # A valid + invalid value should also raise an exception
    def append_invalid_value(value):
        return [*value, 'Invalid_value'] if isinstance(value, list) else [value, 'Invalid_value']

    with pytest.raises(ValueError):
        validate_multiple_choice_params({k: append_invalid_value(v) for k, v in params.items()})


@pytest.mark.parametrize(
    'params, expected_value',
    [
        ({'identifications': 'most agree'}, 'most_agree'),
        ({'interval': 'month of year'}, 'month_of_year'),
    ],
)
def test_validate_multiple_choice_params__normalization(params, expected_value):
    validated_params = validate_multiple_choice_params(params)
    value = next(iter(validated_params.values()))
    assert value == expected_value



from functools import wraps
from logging import getLogger
from math import ceil
from time import sleep
from typing import Callable

from pyinaturalist.constants import (
    EXPORT_URL,
    LARGE_REQUEST_WARNING,
    PER_PAGE_RESULTS,
    REQUESTS_PER_MINUTE,
    THROTTLING_DELAY,
    JsonResponse,
)

logger = getLogger(__name__)


def add_paginate_all(method: str = 'page'):
    """Decorator that adds auto-pagination support, invoked by passing ``page='all'`` to the wrapped
    API function.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **params):
            if params.get('page') == 'all':
                return paginate_all(func, *args, method=method, **params)
            return func(*args, **params)

        return wrapper

    return decorator


def paginate_all(api_func: Callable, *args, method: str = 'page', **params) -> JsonResponse:
    """Get all pages of a multi-page request. Explicit pagination parameters will be overridden.

    Args:
        api_func: API endpoint function to paginate
        method: Pagination method; either 'page', 'id', or 'autocomplete' (see below)
        params: Original request parameters

    Note on pagination by ID, from the iNaturalist documentation:
    _'The large size of the observations index prevents us from supporting the page parameter when
    retrieving records from large result sets. If you need to retrieve large numbers of records,
    use the ``per_page`` and ``id_above`` or ``id_below`` parameters instead.'_

    Returns:
        Response dict containing combined results, in the same format as ``api_func``
    """
    params.pop('page', None)
    if method == 'autocomplete':
        return paginate_autocomplete(api_func, *args, **params)
    if method == 'id':
        params['order_by'] = 'id'
        params['order'] = 'asc'
    else:
        params['page'] = 1
    params['per_page'] = PER_PAGE_RESULTS

    # Run an initial request to get request size
    response = api_func(**params)
    results = page_results = response['results']
    total_results = response.get('total_results')
    estimate_request_size(total_results)

    # Some endpoints (like get_observation_fields) don't return total_results for some reason
    # Also check page size, in case total_results is off (race condition, outdated index, etc.)
    def check_results():
        more_results = total_results is None or len(results) < total_results
        return more_results and len(page_results) > 0

    # Loop until we get all pages
    while check_results():
        if method == 'id':
            params['id_above'] = page_results[-1]['id']
        else:
            params['page'] += 1

        page_results = api_func(**params).get('results', [])
        results += page_results
        sleep(THROTTLING_DELAY)

    return {
        'results': results,
        'total_results': len(results),
    }


def paginate_autocomplete(api_func: Callable, *args, **params) -> JsonResponse:
    """Attempt to get as many results as possible from the places autocomplete endpoint.
    This is necessary for some problematic places for which there are many matches but not ranked
    with the desired match(es) first.

    This works based on different rankings being returned for order_by=area. No other fields can be
    sorted on, and direction can't be specified, but this can at least provide a few additional
    results beyond the limit of 20.
    """
    params['per_page'] = 20
    params.pop('order_by', None)

    # Search with default ordering and ordering by area (if there are more than 20 results)
    page_1 = api_func(*args, **params)
    if page_1['total_results'] > 20:
        page_2 = api_func(*args, **params, order_by='area')
    else:
        page_2 = {'results': []}

    # De-duplicate results
    unique_results = {r['id']: r for page in [page_1, page_2] for r in page['results']}
    return {
        'results': list(unique_results.values()),
        'total_results': page_1['total_results'],
    }


def estimate_request_size(total_results):
    """Log the estimated total number of requests and rate-limiting delay, and show a warning if
    the request is too large
    """
    if not total_results:
        return
    total_requests = ceil(total_results / PER_PAGE_RESULTS)
    est_delay = ceil((total_requests / REQUESTS_PER_MINUTE) * 60)
    logger.info(
        f'This query will fetch {total_results} results in {total_requests} requests. '
        f'Estimated total rate-limiting delay: {est_delay} seconds'
    )

    if total_results > LARGE_REQUEST_WARNING:
        logger.warning(
            'This request is larger than recommended for API usage. For bulk requests, consider '
            f'using the iNat export tool instead: {EXPORT_URL}'
        )



"""
Shared unit test-related utilities.
Pytest will also automatically pick up any fixtures defined here.
"""
import json
import logging
import os
import pytest
import re
from inspect import Parameter, getmembers, isfunction, signature
from os.path import join
from unittest.mock import MagicMock, patch

# If ipdb is installed, register it as the default debugger
from pyinaturalist.constants import SAMPLE_DATA_DIR

try:
    import ipdb  # noqa: F401

    os.environ['PYTHONBREAKPOINT'] = 'ipdb.set_trace'
except ImportError:
    pass


HTTP_FUNC_PATTERN = re.compile(r'(get|put|post|delete)_.+')

MOCK_CREDS_ENV = {
    'INAT_USERNAME': 'valid_username',
    'INAT_PASSWORD': 'valid_password',
    'INAT_APP_ID': 'valid_app_id',
    'INAT_APP_SECRET': 'valid_app_secret',
}
MOCK_CREDS_OAUTH = {
    'username': 'valid_username',
    'password': 'valid_password',
    'client_id': 'valid_app_id',
    'client_secret': 'valid_app_secret',
}

# Enable logging for urllib and other external loggers
logging.basicConfig(level='INFO')
logging.getLogger('pyinaturalist').setLevel('DEBUG')


@pytest.fixture(scope='function', autouse=True)
def patch_ratelimit():
    """Disable rate-limiting during test session"""
    with patch('pyinaturalist.api_requests.ratelimit'):
        yield


def get_module_functions(module):
    """Get all functions belonging to a module (excluding imports)"""
    return {
        name: member
        for name, member in getmembers(module)
        if isfunction(member) and member.__module__ == module.__name__
    }


def get_module_http_functions(module):
    """Get all functions belonging to a module and prefixed with an HTTP method"""
    return {
        name: func
        for name, func in get_module_functions(module).items()
        if HTTP_FUNC_PATTERN.match(name.lower())
    }


def get_mock_args_for_signature(func):
    """Automagically get a list of mock objects corresponding to the required arguments
    in a function's signature. Using ``inspect.Signature``, 'Required' is defined by:
    1. positional and 2. no default
    """
    required_args = [
        p
        for p in signature(func).parameters.values()
        if p.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
        and p.default is Parameter.empty
    ]
    return [MagicMock()] * len(required_args)


def sample_data_path(filename):
    return join(SAMPLE_DATA_DIR, filename)


def load_sample_data(filename):
    with open(sample_data_path(filename), encoding='utf-8') as fh:
        if filename.endswith('json'):
            return json.load(fh)
        else:
            return fh.read()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyinaturalist',
 'pyinaturalist.api_docs',
 'pyinaturalist.models',
 'pyinaturalist.v0',
 'pyinaturalist.v1',
 'pyinaturalist.v2']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.2.0',
 'keyring>=22.3',
 'pyrate-limiter>=2.3.3',
 'python-dateutil>=2.0',
 'python-forge>=18.6.0',
 'requests>=2.20',
 'rich>=10.0']

extras_require = \
{'docs': ['m2r2>=0.2.8',
          'nbsphinx>=0.8.5,<0.9.0',
          'sphinx>=4.0.2,<5.0.0',
          'sphinx-automodapi>=0.13',
          'sphinx-autodoc-typehints>=1.11,<2.0',
          'sphinx-copybutton>=0.3,<0.4',
          'sphinx-material>=0.0.32,<0.0.33',
          'sphinxcontrib-apidoc>=0.3,<0.4']}

setup_kwargs = {
    'name': 'pyinaturalist',
    'version': '0.14.0b0',
    'description': 'iNaturalist API client for python',
    'long_description': '# pyinaturalist\n\n[![Build status](https://github.com/niconoe/pyinaturalist/workflows/Build/badge.svg)](https://github.com/niconoe/pyinaturalist/actions)\n[![Documentation Status (stable)](https://img.shields.io/readthedocs/pyinaturalist/stable?label=docs%20%28main%29)](https://pyinaturalist.readthedocs.io)\n[![Documentation Status (latest)](https://img.shields.io/readthedocs/pyinaturalist/latest?label=docs%20%28dev%29)](https://pyinaturalist.readthedocs.io/en/latest/)\n[![Coverage Status](https://coveralls.io/repos/github/niconoe/pyinaturalist/badge.svg?branch=main)](https://coveralls.io/github/niconoe/pyinaturalist?branch=main)\n\n[![PyPI](https://img.shields.io/pypi/v/pyinaturalist?color=blue)](https://pypi.org/project/pyinaturalist)\n[![PyPI - Python Versions](https://img.shields.io/pypi/pyversions/pyinaturalist)](https://pypi.org/project/pyinaturalist)\n[![PyPI - Format](https://img.shields.io/pypi/format/pyinaturalist?color=blue)](https://pypi.org/project/pyinaturalist)\n[![Run with Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/niconoe/pyinaturalist/main?filepath=examples)\n\n<br/>\n\n[![](docs/images/pyinaturalist_logo_med.png)](https://pyinaturalist.readthedocs.io)\n\n## Summary\n[iNaturalist](https://www.inaturalist.org) is a community science platform that helps people\nget involved in the natural world by observing and identifying the living things around them.\nCollectively, the community produces a rich source of global biodiversity data that can be valuable\nto anyone from hobbyists to scientists.\n\n**pyinaturalist** is an unofficial client for the [iNaturalist API](https://api.inaturalist.org/v1) that\naims to make these data easily accessible in the python programming language. \n\n- [Features](#features)\n- [Quickstart](#quickstart)\n- [Next Steps](#next-steps)\n- [Feedback](#feedback)\n- [Related Projects](#related-projects)\n\n## Features\n* **Easier requests:** Complete type annotations for request parameters, and simplified create/update request formats\n* **Convenient responses:** Type conversions to the things you would expect in python\n* **Docs:** Example requests, responses, scripts, and Jupyter notebooks to help get you started\n* **Security:** Keyring integration for secure credential storage\n* **Server-friendly:** Client-side rate-limiting that follows the\n  [API Recommended Practices](https://www.inaturalist.org/pages/api+recommended+practices),\n  so you can be nice to the iNaturalist servers and not worry about rate-limiting errors\n* **Test-friendly:** A dry-run mode to preview your requests before potentially modifying data\n\nMany of the most relevant API endpoints are included, and more are on the way:\n* **Searching for:**\n    * controlled terms\n    * identifications\n    * observations (multiple formats)\n    * observation fields\n    * observation species counts\n    * places\n    * projects\n    * species\n* **Text search autocompletion for:**\n    * places\n    * species\n    * users\n* **Creating and updating:**\n    * observations\n    * observation fields\n    * observation photos\n    * observation sounds\n\n## Quickstart\nHere are usage examples for some of the most commonly used features.\n\nFirst, install with pip:\n```bash\npip install pyinaturalist\n```\n\nThen, import the main API functions:\n```python\nfrom pyinaturalist import *\n```\n\n### Search observations\nLet\'s start by searching for all your own observations. There are\n[numerous fields you can search on](https://pyinaturalist.readthedocs.io/en/latest/modules/pyinaturalist.node_api.html#pyinaturalist.node_api.get_observations), but we\'ll just use `user_id` for now:\n```python\n>>> observations = get_observations(user_id=\'my_username\')\n```\n\nThe full response will be in JSON format, but we can just print out a few basic details:\n```python\n>>> for obs in observations[\'results\']:\n>>>    print(format_observations(obs))\n[78242978] Species: Agelastica alni (Alder Leaf Beetle) observed by niconoe on 2021-05-10 18:45:38+01:00 at 1428 Braine-l\'Alleud, Belgique\n[78218860] Genus: Bradybatus observed by niconoe on 2021-05-10 15:22:49+01:00 at 1428 Braine-l\'Alleud, Belgique\n...\n```\n\nYou can also get\n[observation counts by species](https://pyinaturalist.readthedocs.io/en/stable/modules/pyinaturalist.node_api.html#pyinaturalist.node_api.get_observation_species_counts).\nOn iNaturalist.org, this information can be found on the \'Species\' tab of search results.\nFor example, to get the counts of all your own research-grade observations:\n```python\n>>> counts = get_observation_species_counts(user_id=\'my_username\', quality_grade=\'research\')\n>>> print(format_species_counts(counts, align=True))\n[48473   ]: Species:          Ganoderma applanatum (Artist\'s bracket): 4\n[50310   ]: Species:         Arisaema triphyllum (Jack-in-the-pulpit): 4\n[50817   ]:   Genus:                     Auricularia (Wood ear fungi): 3\n[81599   ]: Species:                 Silphium perfoliatum (Cup plant): 3\n[120215  ]: Species:    Bombus griseocollis (Brown-belted Bumble Bee): 2\n...\n```\n\nAnother useful format is the\n[observation histogram](https://pyinaturalist.readthedocs.io/en/stable/modules/pyinaturalist.node_api.html#pyinaturalist.node_api.get_observation_histogram),\nwhich shows the number of observations over a given interval. The default is `month_of_year`:\n```python\n>>> histogram = get_observation_histogram(user_id=\'my_username\')\n>>> print(histogram)\n{\n    1: 8,  # January\n    2: 1,  # February\n    3: 19, # March\n    ...,   # etc.\n}\n```\n\n### Create and update observations\nTo create or modify observations, you will first need to log in.\nThis requires creating an [iNaturalist app](https://www.inaturalist.org/oauth/applications/new),\nwhich will be used to get an access token.\n```python\ntoken = get_access_token(\n    username=\'my_username\',\n    password=\'my_password\',\n    app_id=\'my_app_id\',\n    app_secret=\'my_app_secret\',\n)\n```\nSee\n[Authentication](https://pyinaturalist.readthedocs.io/en/latest/user_guide.html#authentication)\nfor additional authentication options, including environment variables, keyrings, and password managers.\n\nNow we can [create a new observation](https://pyinaturalist.readthedocs.io/en/stable/modules/pyinaturalist.rest_api.html#pyinaturalist.rest_api.create_observation):\n```python\nfrom datetime import datetime\n\nresponse = create_observation(\n    taxon_id=54327,  # Vespa Crabro\n    observed_on_string=datetime.now(),\n    time_zone=\'Brussels\',\n    description=\'This is a free text comment for the observation\',\n    tag_list=\'wasp, Belgium\',\n    latitude=50.647143,\n    longitude=4.360216,\n    positional_accuracy=50, # GPS accuracy in meters\n    access_token=token,\n    photos=[\'~/observations/wasp1.jpg\', \'~/observations/wasp2.jpg\']\n)\n\n# Save the new observation ID\nnew_observation_id = response[0][\'id\']\n```\n\nWe can then [update the observation](https://pyinaturalist.readthedocs.io/en/stable/modules/pyinaturalist.rest_api.html#pyinaturalist.rest_api.update_observation) information, photos, or sounds:\n```python\nupdate_observation(\n    17932425,\n    access_token=token,\n    description=\'updated description !\',\n    photos=\'~/observations/wasp_nest.jpg\',\n    sounds=\'~/observations/wasp_nest.mp3\',\n)\n```\n\n### Search species\nLet\'s say you partially remember either a genus or family name that started with **\'vespi\'**-something.\nThe [taxa endpoint](https://pyinaturalist.readthedocs.io/en/stable/modules/pyinaturalist.node_api.html#pyinaturalist.node_api.get_taxa)\ncan be used to search by name, rank, and several other criteria\n```python\n>>> response = get_taxa(q=\'vespi\', rank=[\'genus\', \'family\'])\n```\n\nAs with observations, there is a lot of information in the response, but we\'ll print just a few basic details:\n```python\n>>> print(format_taxa(response))\n[52747] Family: Vespidae (Hornets, Paper Wasps, Potter Wasps, and Allies)\n[92786] Genus: Vespicula\n[84737] Genus: Vespina\n...\n```\n\nOh, that\'s right, it was **\'Vespidae\'**! Now let\'s find all of its subfamilies using its taxon ID\nfrom the results above:\n```python\n>>> response = get_taxa(parent_id=52747)\n>>> print(format_taxa(response))\n[343248] Subfamily: Polistinae (Paper Wasps)\n[ 84738] Subfamily: Vespinae (Hornets and Yellowjackets)\n[119344] Subfamily: Eumeninae (Potter and Mason Wasps)\n...\n```\n\nJust one last example. There is a [taxon autocomplete](https://pyinaturalist.readthedocs.io/en/stable/modules/pyinaturalist.node_api.html#pyinaturalist.node_api.get_taxa_autocomplete)\ntext search endpoint, which is a bit more niche, and is intended for autocomplete interfaces like\nthe one on iNaturalist.org:\n![Taxon autocompletion on iNaturalist.org](docs/images/inat-taxon-autocomplete.png)\n\nBut it also provides an easy way to search the iNaturalist taxonomy database by taxon name.\nHere is a quick example that will run searches from console input:\n\n```python\nwhile True:\n    query = input("> ")\n    response = get_taxa_autocomplete(q=query)\n    print(format_taxa(response, align=True))\n```\n\nExample usage:\n\n```python\n> opilio\n[  527573]        Genus                                              Opilio\n[   47367]        Order                              Opiliones (Harvestmen)\n[   84644]      Species             Phalangium opilio (European Harvestman)\n...\n> coleo\n[  372759]     Subclass     Coleoidea (Octopuses, Squids, and Cuttlefishes)\n[   47208]        Order                                Coleoptera (Beetles)\n[  359229]      Species  Coleotechnites florae (Coleotechnites Flower Moth)\n...\n<Ctrl-C>\n```\n\n## Next Steps\nFor more information, see:\n\n* [User Guide](https://pyinaturalist.readthedocs.io/en/latest/user_guide.html):\n  introduction and general features that apply to most endpoints\n* [Endpoint Summary](https://pyinaturalist.readthedocs.io/en/latest/endpoints.html):\n  a complete list of endpoints wrapped by pyinaturalist\n* [Examples](https://pyinaturalist.readthedocs.io/en/stable/examples.html):\n  data visualizations and other examples of things to do with iNaturalist data\n* [Reference](https://pyinaturalist.readthedocs.io/en/latest/reference.html): Detailed API documentation\n* [Contributing Guide](https://pyinaturalist.readthedocs.io/en/stable/contributing.html):\n  development details for anyone interested in contributing to pyinaturalist\n* [History](https://github.com/niconoe/pyinaturalist/blob/dev/HISTORY.md):\n  details on past and current releases\n* [Issues](https://github.com/niconoe/pyinaturalist/issues): planned & proposed features\n\n## Feedback\nIf you have any problems, suggestions, or questions about pyinaturalist, please let us know!\nJust [create an issue](https://github.com/niconoe/pyinaturalist/issues/new/choose).\nAlso, **PRs are welcome!**\n\n**Note:** pyinaturalist is developed by members of the iNaturalist community, and is not endorsed by\niNaturalist.org or the California Academy of Sciences. If you have non-python-specific questions\nabout iNaturalist, the [iNaturalist Community Forum](https://forum.inaturalist.org/) is the best\nplace to start.\n\n## Related Projects\n* [pyinaturalist-open-data](https://github.com/JWCook/pyinaturalist-open-data):  Tools for working with [iNaturalist open data](https://registry.opendata.aws/inaturalist-open-data/)\n* [pyinaturalist-notebook](https://github.com/JWCook/pyinaturalist-notebook): Jupyter notebook Docker image for pyinaturalist\n* [pyinaturalist-convert](https://github.com/JWCook/pyinaturalist-convert): (WIP) Tools to convert observation data to and from multiple formats\n',
    'author': 'Nicolas Noé',
    'author_email': 'nicolas@niconoe.eu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/niconoe/pyinaturalist',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from dateutil.tz import tzutc

{
    'total_results': 3,
    'page': 1,
    'per_page': 3,
    'results': [
        {
            'id': 886482,
            'login': 'niconoe',
            'spam': False,
            'suspended': False,
            'created_at': datetime(2018, 4, 23, 17, 11, 14, tzinfo=tzutc()),
            'login_autocomplete': 'niconoe',
            'login_exact': 'niconoe',
            'name': 'Nicolas Noé',
            'name_autocomplete': 'Nicolas Noé',
            'orcid': 'https://orcid.org/0000-0002-9503-4750',
            'icon': 'https://static.inaturalist.org/attachments/users/icons/886482/thumb.jpg?1529671435',
            'observations_count': 928,
            'identifications_count': 118,
            'journal_posts_count': 0,
            'activity_count': 1046,
            'species_count': 396,
            'universal_search_rank': 928,
            'roles': [],
            'site_id': 1,
            'icon_url': 'https://static.inaturalist.org/attachments/users/icons/886482/medium.jpg?1529671435',
        },
        {
            'id': 2909130,
            'login': 'niconoerbo',
            'spam': False,
            'suspended': False,
            'created_at': datetime(2020, 5, 5, 6, 28, 32, tzinfo=tzutc()),
            'login_autocomplete': 'niconoerbo',
            'login_exact': 'niconoerbo',
            'name': None,
            'name_autocomplete': None,
            'orcid': None,
            'icon': None,
            'observations_count': 6,
            'identifications_count': 0,
            'journal_posts_count': 0,
            'activity_count': 6,
            'universal_search_rank': 6,
            'roles': [],
            'site_id': 1,
            'icon_url': None,
        },
        {
            'id': 3358478,
            'login': 'nicono',
            'spam': False,
            'suspended': False,
            'created_at': datetime(2020, 7, 20, 18, 7, 44, tzinfo=tzutc()),
            'login_autocomplete': 'nicono',
            'login_exact': 'nicono',
            'name': None,
            'name_autocomplete': None,
            'orcid': None,
            'icon': None,
            'observations_count': 0,
            'identifications_count': 0,
            'journal_posts_count': 0,
            'activity_count': 0,
            'universal_search_rank': 0,
            'roles': [],
            'site_id': 1,
            'icon_url': None,
        },
    ],
}



from functools import update_wrapper
from inspect import signature
from typing import Any, Callable, Iterable, List

from attr import Attribute, Factory

from pyinaturalist.models import BaseModel

FIELD_DEFAULTS = {
    'default': None,
    'validator': None,
    'repr': True,
    'cmp': None,
    'hash': None,
    'init': False,
    'inherited': False,
}


class LazyProperty(property):
    """A lazy-initialized/cached descriptor, similar to ``@functools.cached_property``, but works
    for slotted classes by not relying on ``__dict__``.

    Currently this is used to lazy-load nested model objects for better performance.
    How it works:

    1. Define a `LazyProperty` on a model, say ``MyModel.foo``
    2. Use `add_lazy_attrs` as a field_transformer for the model, which adds an attr.field ``_foo``
    3. During attrs init, ``_foo`` is set from response JSON
    4. When ``foo`` is first called, it converts ``_foo`` from JSON into a model object
    5. When ``foo`` is called again the previously converted ``_foo`` object will be returned

    Example::

        def converter_func(value) -> str:
            return str(value)

        def list_converter_func(value) -> List:
            return [value]

        @attrs.define(field_transformer=add_lazy_attrs)
        class MyModel(BaseModel):
            str_field = LazyProperty(converter_func)
            list_field = LazyProperty(list_converter_func)

            # Auto-generated fields will look like:
            # _str_field = field(factory=list)
            # _list_field = field(default=None)

    """

    def __init__(self, converter: Callable, name: str = None):
        self.converter = converter
        self.default = None
        self.__set_name__(None, name)
        update_wrapper(self, converter)

        # Use either a list factory or default value, depending on the converter's return type
        if _returns_list(converter):
            self.default = Factory(list)

    def __get__(self, obj, cls):
        """When accessing the value, convert it if it hasn't already been, and cache the converted
        value for subsequent calls.
        """
        if obj is None:
            return self

        value = getattr(obj, self.temp_attr)
        if value and not _is_model_object_or_list(value):
            value = self.converter(value)
            setattr(obj, self.temp_attr, value)
        return value

    def __set__(self, obj, raw_value):
        setattr(obj, self.temp_attr, raw_value)

    def __set_name__(self, owner, name):
        self.__name__ = name
        self.temp_attr = f'_{name}'

    def get_lazy_attr(self) -> Attribute:
        """Get an attribute corresponding to this LazyProperty instance"""
        return make_attribute(self.temp_attr, init=True, repr=False, default=self.default)


def add_lazy_attrs(cls, fields):
    """A field transformer to do some post-processing on a model class while it's being created.
    For each :py:class:`.LazyProperty` on a model class, this adds a corresponding ``attr.field``
    in which to temporarily store a raw JSON value that will later be converted into a model object.
    """
    lazy_properties = [p for p in cls.__dict__.values() if isinstance(p, LazyProperty)]
    return list(fields) + [p.get_lazy_attr() for p in lazy_properties]


def get_model_fields(obj: Any) -> Iterable[Attribute]:
    """Add placeholder attributes for lazy-loaded model properties so they get picked up by rich's
    pretty-printer. Does not change behavior for anything except :py:class:`.BaseModel` subclasses.
    """
    from pyinaturalist.models import LazyProperty

    attrs = list(obj.__attrs_attrs__)
    if isinstance(obj, BaseModel):
        prop_names = [k for k, v in type(obj).__dict__.items() if isinstance(v, LazyProperty)]
        attrs += [make_attribute(p) for p in prop_names]
    return attrs


def make_attribute(name, **kwargs):
    kwargs = {**FIELD_DEFAULTS, **kwargs}
    return Attribute(name=name, **kwargs)


def _is_model_object_or_list(value):
    try:
        return isinstance(value, BaseModel) or isinstance(value[0], BaseModel)
    except (AttributeError, KeyError, TypeError):
        return False


def _returns_list(func: Callable) -> bool:
    """Determine if a function is annotated with a List return type"""
    return_type = signature(func).return_annotation
    return _get_origin(return_type) in (list, List)


def _get_origin(tp):
    """Get generic origin (for python 3.6 compatibility)"""
    if getattr(tp, '__origin__', None):
        return tp.__origin__
    if hasattr(tp, '_gorg') and hasattr(tp._gorg, '__mro__'):
        for t in tp._gorg.__mro__:
            if t.__module__ in ('builtins', '__builtin__') and t is not object:
                return t
    return tp
